#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/time.h>

#include "Brain.h"
#include "Eyes.h"
#include "VideoWindow.h"

Brain::Brain(int width, int height) : inactivity_time(0), width(width), height(height), eye_position(-1), eye_direction(0)
{
  video_buffer=(uint8_t *)malloc(width * height * 3);
  diff_buffer=(uint8_t *)malloc(width * height * 3);
  memset(video_buffer, 0, width * height * 3);
  memset(diff_buffer, 0, width * height * 3);
  memset(&eye_move_time, 0, sizeof(eye_move_time));

  eyes = new Eyes();
}

Brain::~Brain()
{
  free(video_buffer);
  free(diff_buffer);
}

void Brain::process(uint8_t *buffer)
{
  int n,i;
  int x,y;
  int value;
  const int THRESHOLD = 50;
  int max = 0;
  int section = -1;

  // Take YUV buffer and subtract it from previous RGB buffer.  This
  // is a little awkward.  The "video_buffer" which is the previous frame
  // is just the Y component of the YUV422 into the G component of the RGB.
  // I think I did this so I could watch the video buffer in green and black.
  i = 0;
  for (n = 0; n < width * height * 3; n = n + 3)
  {
    value = abs((int)buffer[i + 0] - (int)video_buffer[n + 1]);
    value = value > THRESHOLD ? 255 : 0;
    diff_buffer[n + 0] = 0;
    diff_buffer[n + 1] = value;
    diff_buffer[n + 2] = 0;
    value = abs((int)buffer[i + 2] - (int)video_buffer[n + 4]);
    value = value > THRESHOLD ? 255 : 0;
    diff_buffer[n + 3] = 0;
    diff_buffer[n + 4] = value;
    diff_buffer[n + 5] = 0;
    i+=2;
  }

  // Count all the green pixels down the screen and store it in the R
  // component of the RGB diff buffer.
  int bar_width = width / 64;
  int ptr;

  for (x = 0; x < width; x += bar_width)
  {
    int total = 0;

    for (n = 0; n < bar_width; n++)
    {
      //int count=0;
      int ptr = ((x + n) * 3) + 1;
      for (y = 0; y < height; y++)
      {
        if (diff_buffer[ptr] == 255) { total++; }
        ptr += width * 3;
      }
    }

    int avg = total / bar_width;

    if (avg > max && avg > 5) { max = avg; section = x; }

    for (y = 479; y >= 479 - avg; y--)
    {
      ptr = (y * width * 3)+(x * 3);
      for (n = 0; n < bar_width; n++) { diff_buffer[ptr] = 255; ptr += 3; }
    }
  }

  // Copy the Y component from the YUV buffer to the G component of the
  // RGB video_buffer.
  i = 0;
  for (n = 0; n < width * height * 3; n = n + 3)
  {
    video_buffer[n + 1] = buffer[i + 0];
    video_buffer[n + 4] = buffer[i + 2];
    i += 2;
  }

  if (section != -1)
  {
    for (y = 0; y < 50; y++)
    {
      ptr = (y * width * 3) + (section * 3) + 2;
      for (n = 0; n < bar_width; n++) { diff_buffer[ptr] = 255; ptr += 3; }
    }

    // eye_position is the current position of the eyes (0 to 20)
    int new_position = section * 20 / width;
    int curr_direction = new_position > eye_position ? 1 : -1;

    if (curr_direction == eye_direction || can_eyes_change_dir())
    {
      eye_direction = curr_direction;
      eye_position = section * 20 / width;
      eyes->turn(section * 20 / width);
      gettimeofday(&eye_move_time, NULL);
    }

    inactivity_time = 1;
  }
    else
  {
    // If nothing happend for 60 frames (a little more than 2 seconds)
    // turn the eyes to look straight.
    inactivity_time++;
    if (inactivity_time == 30)
    {
      eye_direction = 0;
      eye_position = 10;
      memset(&eye_move_time, 0, sizeof(eye_move_time));

      eyes->turn(10);
    }
  }

  // Convert current to real X position
  section = eye_position * (width / 20);
  for (y = 0; y < 50; y++)
  {
    // Calculate ptr in buffer to start drawing this row for current
    ptr = (y * width * 3) + (section * 3);
    for (n = 0; n < bar_width; n++)
    {
      diff_buffer[ptr] = diff_buffer[ptr + 1] = 255;
      ptr += 3;
    }
  }
}

bool Brain::can_eyes_change_dir()
{
  struct timeval curr_time;
  time_t sec_diff;
  suseconds_t usec_diff;

  if (eye_position <= 1 || eye_position >=19) { return true; }
  if (eye_move_time.tv_sec == 0 && eye_move_time.tv_usec == 0) { return true; }

  gettimeofday(&curr_time, NULL);

  sec_diff = curr_time.tv_sec - eye_move_time.tv_sec;
  if (sec_diff > 1) { sec_diff--; curr_time.tv_usec += 1000000; }
  usec_diff = (sec_diff * 1000000) + curr_time.tv_usec - eye_move_time.tv_usec;

  //printf("usec_diff=%ld\n", usec_diff);

  return usec_diff > 2000000 ? true : false;
}


