#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <asm/types.h>
#include <linux/videodev2.h>
#include <sys/ioctl.h>

#include "Capture.h"

Capture::Capture(int width, int height) : streaming(0), width(width), height(height)
{
  image = (uint8_t *)malloc(width * height * 2);
}

Capture::~Capture()
{
  close(fd);
  free(image);
}

int Capture::init()
{
  fd = open("/dev/video0", O_RDWR, 0);

  if (fd < 0)
  {
    printf("Could not open video device\n");
  }
    else
  {
    struct v4l2_capability vid_cap;
    struct v4l2_format vid_fmt;

    memset(&vid_cap, 0, sizeof(vid_cap));
    memset(&vid_fmt, 0, sizeof(vid_fmt));

    if (ioctl(fd, VIDIOC_QUERYCAP, &vid_cap)==-1)
    {
      printf("ioctl error: VIDIOC_QUERYCAP  errno=%d\n",errno);
      return -1;
    }

    printf("      driver: %16s\n",vid_cap.driver);
    printf("        card: %32s\n",vid_cap.card);
    printf("    bus_info: %32s\n",vid_cap.bus_info);
    printf("     version: %d\n",vid_cap.version);
    printf("capabilities: 0x%08x\n",vid_cap.capabilities);
    printf("\n");

    if ((vid_cap.capabilities&V4L2_CAP_VIDEO_CAPTURE) == 0)
    {
      close(fd);
      printf("This device doesn't support frame capture\n");
      return -3;
    }

    vid_fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vid_fmt.fmt.pix.width = 640;
    vid_fmt.fmt.pix.height = 480;
    //vid_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
    vid_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    //vid_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_UYVY;
    //vid_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_BGR24;
    //vid_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_JPEG;
    //vid_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB32;
    //vid_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV420;
    //vid_fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YVYU;
    //vid_fmt.fmt.pix.field = V4L2_FIELD_NONE;
    vid_fmt.fmt.pix.field = V4L2_FIELD_ANY;
    //vid_fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

    if (ioctl(fd, VIDIOC_S_FMT, &vid_fmt) == -1)
    {
      close(fd);
      perror("VIDIOC_S_FMT");
      //printf("ioctl error: VIDIOC_S_FMT errno=%d\n",errno);
      return -1;
    }

    //memset(&vid_fmt, 0, sizeof(vid_fmt));

    if (ioctl(fd, VIDIOC_G_FMT, &vid_fmt) == -1)
    {
      close(fd);
      printf("ioctl error: VIDIOC_G_FMT errno=%d\n", errno);
      return -1;
    }

    printf("       width: %d\n", vid_fmt.fmt.pix.width);
    printf("      height: %d\n", vid_fmt.fmt.pix.height);
    printf("bytesperline: %d\n", vid_fmt.fmt.pix.bytesperline);
    printf("   sizeimage: %d\n", vid_fmt.fmt.pix.sizeimage);
    printf("      format: %x %c%c%c%c\n",
      vid_fmt.fmt.pix.pixelformat,
      vid_fmt.fmt.pix.pixelformat&0xff,
      (vid_fmt.fmt.pix.pixelformat>>8)&0xff,
      (vid_fmt.fmt.pix.pixelformat>>16)&0xff,
      (vid_fmt.fmt.pix.pixelformat>>24)&0xff
    );

    if ((vid_cap.capabilities&V4L2_CAP_READWRITE) == 0)
    {
      //close(fd);
      printf("This device doesn't support read/write\n");
      //return -3;

      if ((vid_cap.capabilities&V4L2_CAP_STREAMING) == 0)
      {
        //close(fd);
        printf("This device doesn't support read/write\n");
        //return -3;
      }

      streaming=1;

      struct v4l2_requestbuffers reqbuf;
      memset(&reqbuf, 0, sizeof(reqbuf));
      reqbuf.count = 1;
      reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
#if 0
      reqbuf.memory=V4L2_MEMORY_USERPTR;
#endif
      reqbuf.memory=V4L2_MEMORY_MMAP;

      if (ioctl(fd, VIDIOC_REQBUFS, &reqbuf) == -1)
      {
        perror("VIDIO_REQBUFS");
        return -1;
      }

      start_capture();
    }
      else
    {
      image = (uint8_t *)malloc(width * height * 2);
    }

    //window->video_buffer=(unsigned char *)malloc(640*480*3);

    //add_fd(fd, READ, (FileHandler)video_callback, window);
#if 0
    int type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (ioctl(fd, VIDIOC_STREAMON, &type)==-1)
    {
      perror("VIDIOC_STREAMON");
    }
#endif

    //sleep(10);

#if 0
    int len = width * height * 2;
    int n = read(fd, image, len);
    if (n<0)
    {
      perror("v4l2");
      printf("v4l2 read()=%d error.  errno=%d fd=%d image=%p size=%d\n",n,errno,fd,image,len);
    }
#endif
  }

  return 0;
}

void Capture::start_capture_user_ptr()
{
  struct v4l2_buffer buf;
  enum v4l2_buf_type type;

  buffers.length = width * height * 2;
  image = (uint8_t *)malloc(buffers.length);
  buffers.start = image;

  memset(&buf, 0, sizeof(buf));
  type=V4L2_BUF_TYPE_VIDEO_CAPTURE;

  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_USERPTR;
  buf.index = 0;
  buf.m.userptr = (unsigned long)&buffers;
  buf.length = width * height * 2;

  if (ioctl(fd, VIDIOC_QBUF, &buf) == -1)
  {
    perror("VIDIOC_QBUF");
  }

  if (ioctl(fd, VIDIOC_STREAMON, &type) == -1)
  {
    perror("VIDIOC_STREAMON");
  }
}

void Capture::start_capture()
{
  struct v4l2_requestbuffers req;
  struct v4l2_buffer buf;
  enum v4l2_buf_type type;

  printf("start_capture()\n");

  memset(&req, 0, sizeof(req));

  req.count = 1;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;

  if (ioctl(fd, VIDIOC_REQBUFS, &req)==-1)
  {
    perror("VIDIOC_REQBUFS");
  }

  memset(&buf, 0, sizeof(buf));
  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  buf.index = 0;
  //buf.length = width * height * 2;

  if (ioctl(fd, VIDIOC_QUERYBUF, &buf)==-1)
  {
    perror("VIDIOC_QUERYBUF");
  }

  buffers.length = width * height * 2;
  if (buffers.length != buf.length)
  {
    printf("uh oh %s:%d\n", __FILE__, __LINE__);
  }

  buffers.start=mmap(NULL, buffers.length, PROT_READ|PROT_WRITE, MAP_SHARED, fd,buf.m.offset);
  image=(uint8_t *)buffers.start;

  memset(&buf, 0, sizeof(buf));
  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  buf.index = 0;

  if (ioctl(fd, VIDIOC_QBUF, &buf) == -1)
  {
    perror("VIDIOC_QBUF");
  }

  type=V4L2_BUF_TYPE_VIDEO_CAPTURE;

  if (ioctl(fd, VIDIOC_STREAMON, &type)==-1)
  {
    perror("VIDIOC_STREAMON");
  }
}

void Capture::read_frame()
{
  int len;
  int n;

  if (fd > 0)
  {
    if (streaming == 0)
    {
      len = width * height * 2;
      n = read(fd, image, len);
      if (n < 0)
      {
        perror("read()");
        printf("v4l2 read()=%d error.  errno=%d fd=%d image=%p size=%d\n", n, errno, fd, image,len);
      }
    }
      else
    {
      struct v4l2_buffer buf;
      memset(&buf, 0, sizeof(buf));

      buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      buf.memory = V4L2_MEMORY_MMAP;
#if 0
      buf.memory = V4L2_MEMORY_USERPTR;
#endif

      if (ioctl(fd, VIDIOC_DQBUF, &buf) == -1)
      {
        perror("VIDIOC_DQBUF");
      }

      //window->add_frame(image);
      //window->redraw();

      if (ioctl(fd, VIDIOC_QBUF, &buf) == -1)
      {
        perror("VIDIOC_DQBUF");
      }
    }
  }

  //convert_yuyv(window->video_buffer, image, 640, 480);
  //repeat_timeout(0.033, timeout_callback, window);
}

